<h4 style="text-align: center">Percent of Ensembles that Exceed Return Periods</h4>
<table id="returntable" align="center">
  <tbody>
    <tr>
      <th>Dates</th>
      {% for day in days %}
        <th>{{ day }}</th>
      {% endfor %}
    </tr>
    <tr>
      <td>2-yr Return Period</td>
      {% for r in r2 %}
        {% if r != 0 %}
          <td style="background-color: {{ colors['2 Year']}}">{{ r }}%</td>
        {% else %}
          <td>-</td>
        {% endif %}
      {% endfor %}
    </tr>
    <tr>
      <td>5-yr Return Period</td>
      {% for r in r5 %}
        {% if r != 0 %}
          <td style="background-color: {{ colors['5 Year']}}">{{ r }}%</td>
        {% else %}
          <td>-</td>
        {% endif %}
      {% endfor %}
    </tr>
    <tr>
      <td>10-yr Return Period</td>
      {% for r in r10 %}
        {% if r != 0 %}
          <td style="background-color: {{ colors['10 Year']}}">{{ r }}%</td>
        {% else %}
          <td>-</td>
        {% endif %}
      {% endfor %}
    </tr>
    <tr>
      <td>25-yr Return Period</td>
      {% for r in r25 %}
        {% if r != 0 %}
          <td style="background-color: {{ colors['25 Year']}}">{{ r }}%</td>
        {% else %}
          <td>-</td>
        {% endif %}
      {% endfor %}
    </tr>
    <tr>
      <td>50-yr Return Period</td>
      {% for r in r50 %}
        {% if r != 0 %}
          <td style="background-color: {{ colors['50 Year']}}">{{ r }}%</td>
        {% else %}
          <td>-</td>
        {% endif %}
      {% endfor %}
    </tr>
    <tr>
      <td>100-yr Return Period</td>
      {% for r in r100 %}
        {% if r != 0 %}
          <td style="background-color: {{ colors['100 Year']}}">{{ r }}%</td>
        {% else %}
          <td>-</td>
        {% endif %}
      {% endfor %}
    </tr>
  </tbody>
</table>